// Copyright (C) 2004, 2015 Christian Jaekl

// Abstract class representing the result of parsing an XML Element.
// A class derived from this one will know how to parse a subtree inside an XML file, and 
// will contain the result of that parse within itself when the parse has completed.
//
// Note that this code will need to be augmented and fixed if XML namespace support is desired.

package net.jaekl.qd.xml;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;

import org.xml.sax.Attributes;
import org.xml.sax.helpers.AttributesImpl;

public abstract class ParseResult
{
	public static final String TAG_FIELD_NAME = "TAG";
	
	Stack<CurrentInfo> m_current;								// Name of the element that we're currently inside
	StringBuilder m_chars;										// character content of m_current.peek()
	ArrayList<ParseResult> m_childParsers;						// Set of all child parsers
	boolean m_haveSeenMyTag;									// Have I encountered my own (root) tag yet?

	String m_tagName;											// Name of the (root) element tag that I'm parsing
	HashSet<String> m_internal;									// Tags that we will store as members of this class instance
	HashMap<String,Class<? extends ParseResult>> m_external;	// Tags that we will store as child ParseResult-derived objects

	// Information about the "current" tag, which we'll keep on the m_current stack.
	static class CurrentInfo {
		String m_tagName;
		Attributes m_attr;
		
		public CurrentInfo(String tagName, Attributes attr) {
			m_tagName = tagName;
			m_attr = new AttributesImpl(attr);
		}
		
		public String getTagName() { return m_tagName; }
		public final Attributes getAttributes() { return m_attr; }
	}
	
	@SuppressWarnings("unchecked")
	public ParseResult(String tagName, String[] internalMemberTags, Object[][] externalParserTags)
	{
		m_current = new Stack<CurrentInfo>();
		m_chars = new StringBuilder();
		m_childParsers = new ArrayList<ParseResult>();
		m_haveSeenMyTag = false;
		
		m_tagName = tagName;
		m_internal = new HashSet<String>();
		m_external = new HashMap<String, Class<? extends ParseResult>>();

		for (String internalTag : internalMemberTags) {
			m_internal.add(internalTag);
		}

		for (int idx = 0; idx < externalParserTags.length; ++idx) {
			String externalTag = (String)externalParserTags[idx][0];
			Class<? extends ParseResult>  parserClass = (Class<? extends ParseResult>)externalParserTags[idx][1];
			m_external.put(externalTag, parserClass);
		}
	}

	public abstract void endContents(String uri, String localName, String qName, String chars) throws XmlParseException;
	public abstract void endExternal(String uri, String localName, String qName) throws XmlParseException;

	// Called once for this tag itself
	public void handleMainAttributes(Attributes attr) throws MissingAttributeException {
		// Default action is to do nothing.
		// Override this if you want to process tag attributes.
	}
	
	// Called once for each internally-handled subtag
	public void handleInternalAttributes(String tagName, Attributes attr) throws MissingAttributeException {
		// Default action is to do nothing.
		// Override this if you want to process tag attributes.
	}
	
	public String getTagName() { return m_tagName; }

	public void characters(char[] ch, int start, int length) throws XmlParseException
	{
		m_chars.append(ch, start, length);
	}
	
	// Dump human-readable text describing this tag and its children.
	// Default implemenation:  print the tag name (only)
	public void dump(PrintWriter pw, int indent) {
		String margin = String.format("%"+indent+"s", "");
		pw.println(margin + getTagName());
	}
	
	protected ParseResult[] collectParsedChildren(Class<? extends ParseResult> cls) {
		ArrayList<ParseResult> collection = new ArrayList<ParseResult>();
		Iterator<ParseResult> iter = m_childParsers.iterator();
		while (iter.hasNext()) {
			ParseResult pr = iter.next();
			if (pr.getClass().isAssignableFrom(cls)) {
				collection.add(pr);
				iter.remove();
			}
		}
		
		ParseResult[] result = new ParseResult[collection.size()];
		return collection.toArray(result);
	}
	
	// returns true if this ParseResult's context has ended with this endElement() call
	protected boolean endElement(String uri, String localName, String qName) throws XmlParseException
	{
		assert (null != localName);
		
		boolean isInternal = m_internal.contains(localName);

		if (! m_haveSeenMyTag) {
			// We're in some unrecognised prologue.  Ignore it and move on.
			return false;
		}
		
		if (m_tagName.equals(localName)) {
			complete();
			return true;
		}
		
		if (!isInternal) {
			// Unrecognized tag.  Ignore it.
			return false;
		}
		
		CurrentInfo info = m_current.pop();
		String tag = info.getTagName();
		if ( ! tag.equals(localName) ) {
			throw new MismatchedTagsException(tag, localName);
		}
		
		String chars = m_chars.toString();
		endContents(uri, localName, qName, chars);
		
		return false;
	}
	
	// returns either itself, or a new ParseResult-derived object, whichever should handle parsing the inside of this element
	public ParseResult startElement(String uri, String localName, String qName, Attributes attrs) 
			throws XmlParseException
	{
		assert (null != localName);

		m_chars.setLength(0);
		
		if (! m_haveSeenMyTag) {
			// Have we opened our own (root) tag yet?
			
			if (m_tagName.equals(localName)) {
				m_haveSeenMyTag = true;
				handleMainAttributes(attrs);
				return this;
			}
			else {
				// One of two things has happened here.
				// Either (a) we've got some sort of wrapper here, and have not yet reach our own tag, 
				//     or (b) we're parsing XML that doesn't match expectations.
				// In either case, we're going to ignore this tag, and scan forward looking for our own root.
				return this;
			}
		}

		if (m_internal.contains(localName)) {
			handleInternalAttributes(localName, attrs);
			CurrentInfo info = new CurrentInfo(localName, attrs);
			m_current.push(info);
			return this;
		}

		Class<? extends ParseResult> parserClass = m_external.get(localName);
		if (null != parserClass) {
			try {
				ParseResult childParser = (ParseResult) parserClass.newInstance();
				m_childParsers.add(childParser);
				return childParser.startElement(uri, localName, qName, attrs);
			}
			catch (IllegalAccessException iae) {
				throw new XmlParseException(iae);
			}
			catch (InstantiationException ie) {
				throw new XmlParseException(ie);
			}
		}
		
		// Not a recognized tag.  Ignore it, rather than complaining. 
		return this;
	}
	
	public void complete() throws XmlParseException
	{
		// Default implementation is a no-op.
		// Override if you want to validate on endElement()
	}
	
	protected String getRequiredAttr(String tagName, Attributes attr, String attrName)
		throws MissingAttributeException
	{
		String result = attr.getValue(attrName);
		if (null == result) {
			throw new MissingAttributeException(tagName, attrName);
		}
		return result;
	}
	
	protected String getOptionalAttr(Attributes attr, String attrName, String defaultValue)
	{
		String value = attr.getValue(attrName);
		if (null == value) {
			return defaultValue;
		}
		return value;
	}

}

